#ifndef MMSCHD_HH
#define MMSCHD_HH
#include <click/element.hh>
#include <click/hashtable.hh>
#include <click/ipaddress.hh>
#include <click/ipflowid.hh>
#include <click/etheraddress.hh>
#include <click/timer.hh>
#include <clicknet/ether.h>
#include <click/straccum.hh>

CLICK_DECLS

class Spinlock;

class FlowControlBlock{  
   public:
   uint32_t _numpkts;
   Packet *_head;
   Packet *_tail;
   IPAddress dstip;
   IPAddress srcip;
   uint16_t sport;
   uint16_t dport;
   Timestamp _last_activity_time;
   
   FlowControlBlock() : _numpkts(0), _head(0), _tail(0), dstip(0), srcip(0), sport(0), dport(0), _last_activity_time(0)  { }
};

typedef class FlowControlBlock FCB;

struct control_queue{  
   public:
   uint32_t _numpkts;
   Packet *_head;
   Packet *_tail;
   control_queue() : _numpkts(0), _head(0), _tail(0) { }
};

class MMScheduler : public Element{ public:
  MMScheduler();
  ~MMScheduler();

  /*  Input ports details:- port 1: IP pkts port 2: Control packets ex. ARP Requests etc..
     Output ports details: port 1: pull port, to send all pkts to the connected NIC 
  */

  const char *class_name() const	{ return "MMScheduler"; }
  const char *port_count() const 	{ return "1-2/1"; }
  const char *processing() const 	{ return "h/l"; }

  int configure(Vector<String> &conf, ErrorHandler *errh);
  int initialize(ErrorHandler *);
  void cleanup(CleanupStage);
  void push(int port, Packet *p);
  void append_packet(Packet *p);
  Packet* schd_packet();
  Packet* pull(int port);
  FCB* find(IPAddress, uint16_t, IPAddress, uint16_t);
  void add_handlers();  

private: 
  ErrorHandler *_errh;
  static String read_handler(Element *, void *);
  uint8_t _numflows;
  typedef HashTable<IPFlowID, FCB> FlowTable;
  typedef FlowTable::const_iterator FIter;
  FlowTable _flows;
  IPFlowID _next;
  Spinlock* _lock;    

  struct control_queue _ctlq;
  struct control_queue _datq;
  uint8_t _size_click_ip;
  uint8_t _size_click_ether;
  uint8_t _size_eth_ip;

  static int write_handler(const String &s, Element *e, void *vparam, ErrorHandler *errh);
  static void static_flow_mgmt_hook(Timer *, void *e) { ((MMScheduler *) e)->flow_mgmt_hook(); }
  Timer _flow_mgmt_timer;
  void flow_mgmt_hook();
  uint32_t _max_idle_time;
  uint32_t _phy_rate;

  typedef HashTable<uint8_t, uint8_t> RateTable;
  typedef RateTable::const_iterator RIter;
  RateTable _rates;

};
CLICK_ENDDECLS
#endif
